@@ -7,5 +7,5 @@ import pathlib
 import ROOT
 ROOT.PyConfig.IgnoreCommandLineOptions = True
 macro_path = pathlib.Path(__file__).parent.absolute()
-ROOT.gSystem.Load(os.path.join(macro_path, 'macros', 'RooTwoSidedCBShape_cc'))
+ROOT.gSystem.Load(os.path.join(macro_path, 'macros', 'RooTwoSidedCBShape_cxx'))
 from ROOT import RooTwoSidedCBShape
\ No newline at end of file
