@@ -100,8 +100,25 @@ class ExtendedModel(object):
             if (component.InheritsFrom(ROOT.RooAbsPdf.Class()) and condition(name)):
                 component.setAttribute("NOCacheAndTrack")
                 print('INFO: Deactivated level 2 constant term optimizatio for {}'.format(name))
-                                     
-                                
+    
+    @staticmethod
+    def get_nuisance_parameter_names(fname:str, ws_name:str, model_config_name:str):
+        if not os.path.exists(fname):
+            raise FileNotFoundError('workspace file {} does not exist'.format(fname))
+        file = ROOT.TFile(fname) 
+        if (not file):
+            raise RuntimeError("Something went wrong while loading the root file: {}".format(fname))
+        ws = file.Get(ws_name)
+        if not ws:
+            raise RuntimeError("Something went wrong while loading the workspace: {}".format(ws_name))
+        model_config = ws.obj(model_config_name)
+        if not model_config:
+            raise RuntimeError('ERROR: Failed to load model config {}'.format(model_config_name))
+        nuisance_parameters = model_config.GetNuisanceParameters()
+        if not nuisance_parameters:
+            raise RuntimeError('ERROR: Failed to load nuisance parameters')
+        return [nuis.GetName() for nuis in nuisance_parameters]
+            
     def initialize(self):
         if not os.path.exists(self.fname):
             raise FileNotFoundError('workspace file {} does not exist'.format(self.fname))
