@@ -21,4 +21,9 @@ def reorder_arrays(*arrays, descending:bool=True):
     else:
         indices = np.argsort(arrays[0])
     for arr in arrays:
-        arr[:] = arr[indices]    
\ No newline at end of file
+        arr[:] = arr[indices]    
+        
+        
+def reverse_arrays(*arrays):
+    for arr in arrays:
+        arr[:] = arr[::-1] 
\ No newline at end of file
